#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DS1_MAX_FLOOR  2
#define DS1_MAX_WALL   4
#define DS1_MAX_SHAD   1
#define DS1_MAX_SUBST  1

#define FALSE  0
#define TRUE  -1

typedef struct
{
   // bits 0-7, contain layer drawing priority (but not only ?)
   unsigned char prop1;

   // bits 8-13, range from 0 to 63
   unsigned char sub_index;

   // bits 14-19, unknown, at least bits 16 & 17 used sometimes
   unsigned char unk1;

   // bits 20-25, range from 0 to 63
   unsigned char main_index;

   // bits 26-30, unknown, at least bits 26 to 29 used sometimes
   unsigned char unk2;

   // bit 31
   unsigned char hidden;
} FLOOR_S, SHAD_S;

typedef struct
{
   // bits 0-7 of orientation layer
   unsigned char orientation;

   // bits 8-31 of direction layer, should be 0
   unsigned long zero;

   // bits 0-7, contain layer drawing priority (but not only ?)
   unsigned char prop1;

   // bits 8-13, range from 0 to 63
   unsigned char sub_index;

   // bits 14-19, unknown, at least bits 16 & 17 used sometimes
   unsigned char unk1;

   // bits 20-25, range from 0 to 63
   unsigned char main_index;

   // bits 26-30, unknown, at least bits 26 to 29 used sometimes
   unsigned char unk2;

   // bit 31
   unsigned char hidden;
} WALL_S;

typedef struct
{
   // unknown
   unsigned long data;
} SUBST_S;

typedef struct
{
   // floor layers of this tile
   FLOOR_S floor [DS1_MAX_FLOOR];

   // wall layers of this tile
   WALL_S wall [DS1_MAX_WALL];

   // shadow layer of this tile
   SHAD_S shadow [DS1_MAX_SHAD];

   // (optional) substitution layer of this tile
   SUBST_S substitution [DS1_MAX_SUBST];
} TILE_S;

typedef struct PATH_S
{
   long x;
   long y;
   long action;
   struct PATH_S * prev_path;
   struct PATH_S * next_path;
} PATH_S;

typedef struct OBJECT_S
{
   // datas of this object

   long type;
   long id;
   long x;
   long y;
   long flags;

   // necessary for list of objects

   struct OBJECT_S * prev_object;
   struct OBJECT_S * next_object;

   // paths

   long          path_num;
   struct PATH_S * first_path;
} OBJECT_S;

typedef struct
{
   // upper / left corner X coordinate of group, in tiles
   long x0;

   // upper / left corner Y coordinate of group, in tiles
   long y0;

   // group width, in # of tiles
   long width;

   // group height, in # of tiles
   long height;

   // still unknown
   long unknown;
} SUBST_GROUP_S;

typedef struct
{
   // readed version, from 1 to 18
   long version;

   // width of map, in # of tiles
   long width;

   // height of map, in # of tiles
   long height;

   // original width of map, in # of tiles, before resizing
   long original_width;

   // original height of map, in # of tiles, before resizing
   long original_height;

   // act, from 1 to 5
   long act;

   // 0 if none, else 1 or 2
   long substitution_type;

   // # of file strings
   long file_num;

   // table of file string pointers
   char ** file_ptr;

   // number of wall & orientation layers used
   int  wall_num;

   // number of floor layers used
   int  floor_num;

   // number of shadow layer used
   int  shadow_num;

   // number of substitution layer used
   int  substitution_num;  

   // number of substitution goups, datas between objects & NPC paths
   long substitution_groups_num;

   // table of substitution groups datas
   SUBST_GROUP_S * group;

   // map's tiles buffer
   TILE_S * tile_buffer;

   // table of pointers on map's tile start of line
   // usage : ds1->tile[y][x].wall[n].main_index = m;
   TILE_S ** tile;

   // # of objects
   int objects_num;

   // objects are in a linked list, here's the start
   OBJECT_S * first_object;
} DS1_S;

typedef enum
{
   NONE,                   // still no error
   DISK_LOAD_FAILED,       // fopen(ds1name) failed
   BUFFER_MALLOC_FAILED,   // malloc(size of ds1 file) failed
   BUFFER_NULL,            // can't use a buffer pointing to NULL
   DS1_MALLOC_FAILED,      // malloc() of sizeof(DS1_S) failed

   DS1_EOF,                // unexpected Enf Of File
                           // note : this doesn't free() ressources, so you
                           // can try to use what was readed

   FILE_PTR_MALLOC_FAILED, // file_ptr malloc() failed
   FILE_STR_MALLOC_FAILED, // file_ptr[n] malloc() failed
   TILES_MALLOC_FAILED,    // map's tiles malloc() failed
   OBJ_MALLOC_FAILED,      // malloc() of 1 object failed
   GROUPS_MALLOC_FAILED,   // malloc() of substitution groups failed
   PATH_MALLOC_FAILED      // malloc() of 1 path failed
} DS1_ERROR_CODE;

// global
DS1_ERROR_CODE glb_ds1_error = NONE;



// ====================================================================================
// load ds1 file from disk into memory
//
//    name          = path & name of the file to load in mem
//    buffer_length = pointer on the var that will have the buffer length
//
//    return a pointer to the file in mem, or NULL if error
//       (check glb_ds1_error in this case)
//    ---> up to you to free() that buffer
// ====================================================================================
char * ds1_disk_to_mem(char * ds1name, long * buffer_length)
{
   FILE * in;
   char * buffer;


   // init
   * buffer_length = 0;

   // open file
   in = fopen(ds1name, "rb");
   if (in == NULL)
   {
      glb_ds1_error = DISK_LOAD_FAILED;
      return NULL;
   }

   // size of file
   fseek(in, 0, SEEK_END);
   * buffer_length = ftell(in);
   fseek(in, 0, SEEK_SET);

   // malloc
   buffer = (char *) malloc(* buffer_length);
   if (buffer == NULL)
   {
      fclose(in);
      glb_ds1_error = BUFFER_MALLOC_FAILED;
      return NULL;
   }

   // copy file
   fread(buffer, * buffer_length, 1, in);

   // end
   fclose(in);
   return buffer;
}


// ====================================================================================
// release ALL reserved memory from that ds1
// ====================================================================================
void ds1_destroy (DS1_S * ds1)
{
   OBJECT_S * curr_obj,  * next_obj;
   PATH_S   * curr_path, * next_path;
   int      i;


   // for safety
   if (ds1 == NULL)
      return;

   // file_ptr
   if (ds1->file_ptr != NULL)
   {
      for (i=0; i < ds1->file_num; i++)
         free(ds1->file_ptr[i]);
      free(ds1->file_ptr);
   }

   // tile buffer
   if (ds1->tile_buffer != NULL)
      free(ds1->tile_buffer);

   // tile start of line
   if (ds1->tile != NULL)
      free(ds1->tile);

   // object & path
   curr_obj = ds1->first_object;
   while (curr_obj != NULL)
   {      
      curr_path = curr_obj->first_path;
      while (curr_path != NULL)
      {
         next_path = curr_path->next_path;
         free(curr_path);
         curr_path = next_path;
      }
      next_obj = curr_obj->next_object;
      free(curr_obj);
      curr_obj = next_obj;
   }

   // substitution groups
   if (ds1->group != NULL)
      free(ds1->group);

   // ds1
   free(ds1);
}


// ====================================================================================
// read a ds1 file (from mem) and fill a DS1_S structure
//
//    buffer        = pointer on the ds1 file in mem
//    buffer_length = size of this buffer
//    new_width     = for resizing a map, set it to -1 for no resizing
//    new_height    = for resizing a map, set it to -1 for no resizing
//
//    return a pointer on a DS1_S structure, or NULL if error
//       (check glb_ds1_error in this case)
//    ---> up to you to ds1_destroy() this DS1_S structure
//
//    note that despite a pointer is returned, an EOF error could have occured,
//    so check the ds1 error code, to see if it's set to DS1_EOF or not
//
//    you can free() the ds1 buffer if you want, after that function, as NO
//    datas of the DS1_S structure are pointing to it
// ====================================================================================
DS1_S * ds1_load(char * buffer, long buffer_length, int new_width, int new_height)
{
   unsigned long dw;
   DS1_S         * ds1;
   OBJECT_S      obj,  * curr_obj,  * save_obj;
   PATH_S        path, * curr_path, * save_path;
   long          * lptr = (long *) buffer,
                 remain_bytes = buffer_length, // each time we read datas,
                                               // it's decremented
                 new_obj_num, npc_num, path_num, new_path_num, lx, ly, n;
   int           w_num, f_num, s_num, t_num, found;
   int           i, s, x, y, c, max_x, max_y, max_subtile_x, max_subtile_y;
   char          * cptr;
   enum          {W1, W2, W3, W4, O1, O2, O3, O4, F1, F2, SHAD, SUB} lay_stream[14];
   int           dir_lookup[25] = { // direction conversion table
                     0x00, 0x01, 0x02, 0x01, 0x02, 0x03, 0x03, 0x05, 0x05, 0x06,
                     0x06, 0x07, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E,
                     0x0F, 0x10, 0x11, 0x12, 0x14
                 },
                 nb_layer;


   // safety check
   if (buffer == NULL)
   {
      glb_ds1_error = BUFFER_NULL;
      return NULL;
   }

   // ds1 malloc
   ds1 = (DS1_S *) malloc(sizeof(DS1_S));
   if (ds1 == NULL)
   {
      glb_ds1_error = DS1_MALLOC_FAILED;
      return NULL;
   }

   // init this buffer to all zeros
   memset(ds1, 0, sizeof(DS1_S));

   // other inits
   w_num = 0; // # of wall & orientation layers, 0 to 4
   f_num = 0; // # of floor layers,              0 to 2
   s_num = 1; // # of shadow layers,             always 1
   t_num = 0; // # of substitution layer,        0 or 1
   
   // read version
   if (remain_bytes >= 4)
   {
      ds1->version = * lptr;
      lptr++;
      remain_bytes -= 4;
   }
   else
   {
      glb_ds1_error = DS1_EOF;
      return ds1;
   }

   // read width
   if (remain_bytes >= 4)
   {
      ds1->width = ds1->original_width = (* lptr) + 1;
      lptr++;
      remain_bytes -= 4;
   }
   else
   {
      glb_ds1_error = DS1_EOF;
      return ds1;
   }

   // read height
   if (remain_bytes >= 4)
   {
      ds1->height = ds1->original_height = (* lptr) + 1;
      lptr++;
      remain_bytes -= 4;
   }
   else
   {
      glb_ds1_error = DS1_EOF;
      return ds1;
   }

   // read (optional) act
   ds1->act = 1;
   if (ds1->version >= 8)
   {
      if (remain_bytes >= 4)
      {
         ds1->act = (* lptr) + 1;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }
      if (ds1->act > 5)
         ds1->act = 5;
   }

   // read (optional) substitution layer type
   ds1->substitution_type = 0;
   if (ds1->version >= 10)
   {
      if (remain_bytes >= 4)
      {
         ds1->substitution_type = * lptr;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }

      // adjust eventually the # of substitution layer
      if ((ds1->substitution_type == 1) || (ds1->substitution_type == 2))
         t_num = 1;
   }

   // file name strings (useless)
   ds1->file_num = 0;
   ds1->file_ptr = NULL;
   if (ds1->version >= 3)
   {
      if (remain_bytes >= 4)
      {
         ds1->file_num = * lptr;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }

      // table of pointer
      ds1->file_ptr = (char **) malloc(sizeof(char *) * ds1->file_num);
      if (ds1->file_ptr == NULL)
      {
         glb_ds1_error = FILE_PTR_MALLOC_FAILED;
         return ds1;
      }
      memset(ds1->file_ptr, 0, ds1->file_num);

      // fill table of pointer
      cptr = (char *) lptr;
      for (i=0; i < ds1->file_num; i++)
      {
         s = strlen(cptr) + 1;
         if (s <= remain_bytes)
         {
            // malloc for this string
            ds1->file_ptr[i] = (char *) malloc(s);
            if (ds1->file_ptr[i] == NULL)
            {
               glb_ds1_error = FILE_STR_MALLOC_FAILED;
               return ds1;
            }

            // copy it, termination zero included
            strncpy(ds1->file_ptr[i], cptr, s);

            // next string
            cptr += s; // also skip termination zero
            remain_bytes -= s;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }
      }
      lptr = (long *) cptr;

   }

   // skip 2 (optional, and meaningless) dwords
   if ((ds1->version >= 9) && (ds1->version <= 13))
   {
      lptr += 2;
      remain_bytes -= 8;
   }

   // (optional) number of wall & floor layers
   if (ds1->version >= 4)
   {
      // read number of wall (and orientation) layerss
      if (remain_bytes >= 4)
      {
         w_num = * lptr;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }

      // read number of floor layers
      if (ds1->version >= 16)
      {
         if (remain_bytes >= 4)
         {
            f_num= * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }
      }
      else
         f_num = 1; // default # of floor layer
   }
   else // in version < 4
   {
      // # of layers hardcoded
      w_num = 1;
      f_num = 1;
      t_num = 1;
   }

   // copy these values
   ds1->wall_num         = w_num;
   ds1->floor_num        = f_num;
   ds1->shadow_num       = s_num;
   ds1->substitution_num = t_num;

   // which order in file ?
   if (ds1->version < 4)
   {
      lay_stream[0] = W1;   // wall 1
      lay_stream[1] = F1;   // floor 1
      lay_stream[2] = O1;   // orientation 1
      lay_stream[3] = SUB;  // substitution
      lay_stream[4] = SHAD; // shadow
      nb_layer = 5;
   }
   else
   {
      nb_layer = 0;
      for (i=0; i < w_num; i++)
      {
         lay_stream[nb_layer++] = W1 + i; // wall 1 to 4
         lay_stream[nb_layer++] = O1 + i; // orientation 1 to 4
      }
      for (i=0; i < f_num; i++)
         lay_stream[nb_layer++] = F1 + i; // floor 1 to 2
      if (s_num)
         lay_stream[nb_layer++] = SHAD;   // shadow
      if (t_num)
         lay_stream[nb_layer++] = SUB;    // substitution
   }

   // allocate map's tiles
   if ((new_width == -1) || (new_height == -1))
   {
      s     = sizeof(TILE_S) * ds1->width * ds1->height;
      max_x = ds1->width;
      max_y = ds1->height;
   }
   else
   {
      s     = sizeof(TILE_S) * new_width * new_height;
      max_x = new_width;
      max_y = new_height;
   }
   ds1->tile_buffer = (TILE_S *) malloc(s);
   if (ds1->tile_buffer == NULL)
   {
      glb_ds1_error = TILES_MALLOC_FAILED;
      return ds1;
   }
   memset(ds1->tile_buffer, 0, s);

   // allocate map start of line
   s = sizeof(TILE_S *) * max_y;
   ds1->tile = (TILE_S **) malloc(s);
   if (ds1->tile == NULL)
   {
      glb_ds1_error = TILES_MALLOC_FAILED;
      return ds1;
   }

   // fill start of line
   for (y=0; y < max_y; y++)
      ds1->tile[y] = & ds1->tile_buffer[y * max_x];
   
   // read tiles of all layers
   for (i=0; i < nb_layer; i++)
   {
      for (y=0; y < ds1->height; y++)
      {
         for (x=0; x < ds1->width; x++)
         {
            if (remain_bytes >= 4)
            {
               dw = * lptr;
               lptr++;
               remain_bytes -= 4;
               if ((x < max_x) && (y < max_y))
               {
                  // this tile is in the map, even after a resizing
                  switch (lay_stream[i])
                  {
                     // walls
                     case  W1:
                     case  W2:
                     case  W3:
                     case  W4:
                        s = lay_stream[i] - W1; // wall layer index, from 0 to 3
                        ds1->tile[y][x].wall[s].prop1      = (unsigned char) (dw & 0x000000FFLU);
                        ds1->tile[y][x].wall[s].sub_index  = (unsigned char) ((dw & 0x00003F00LU) >> 8);
                        ds1->tile[y][x].wall[s].unk1       = (unsigned char) ((dw & 0x000FC000LU) >> 14);
                        ds1->tile[y][x].wall[s].main_index = (unsigned char) ((dw & 0x03F00000LU) >> 20);
                        ds1->tile[y][x].wall[s].unk2       = (unsigned char) ((dw & 0x7C000000LU) >> 26);
                        ds1->tile[y][x].wall[s].hidden     = (unsigned char) ((dw & 0x80000000LU) >> 31);
                        break;

                     // walls orientation
                     case  O1:
                     case  O2:
                     case  O3:
                     case  O4:
                        s = lay_stream[i] - O1; // orientation layer index, from 0 to 3
                        c = (int) (dw & 0x000000FFLU);
                        if (ds1->version < 7) // convert direction in old ds1
                        {
                           if (c < 25) // just for safety
                              c = dir_lookup[c];
                        }
                        ds1->tile[y][x].wall[s].orientation = (unsigned char) c;
                        ds1->tile[y][x].wall[s].zero        = (unsigned char) ((dw & 0xFFFFFF00LU) >> 8);
                        break;

                     // floors
                     case  F1:
                     case  F2:
                        s = lay_stream[i] - F1; // floor layer index, from 0 to 1
                        ds1->tile[y][x].floor[s].prop1      = (unsigned char) (dw & 0x000000FFLU);
                        ds1->tile[y][x].floor[s].sub_index  = (unsigned char) ((dw & 0x00003F00LU) >> 8);
                        ds1->tile[y][x].floor[s].unk1       = (unsigned char) ((dw & 0x000FC000LU) >> 14);
                        ds1->tile[y][x].floor[s].main_index = (unsigned char) ((dw & 0x03F00000LU) >> 20);
                        ds1->tile[y][x].floor[s].unk2       = (unsigned char) ((dw & 0x7C000000LU) >> 26);
                        ds1->tile[y][x].floor[s].hidden     = (unsigned char) ((dw & 0x80000000LU) >> 31);
                        break;

                     // shadow
                     case  SHAD:
                        s = 0; // shadow layer index
                        ds1->tile[y][x].shadow[s].prop1      = (unsigned char) (dw & 0x000000FFLU);
                        ds1->tile[y][x].shadow[s].sub_index  = (unsigned char) ((dw & 0x00003F00LU) >> 8);
                        ds1->tile[y][x].shadow[s].unk1       = (unsigned char) ((dw & 0x000FC000LU) >> 14);
                        ds1->tile[y][x].shadow[s].main_index = (unsigned char) ((dw & 0x03F00000LU) >> 20);
                        ds1->tile[y][x].shadow[s].unk2       = (unsigned char) ((dw & 0x7C000000LU) >> 26);
                        ds1->tile[y][x].shadow[s].hidden     = (unsigned char) ((dw & 0x80000000LU) >> 31);
                        break;

                     // substitution
                     case  SUB:
                        s = 0; // substitution layer index
                        ds1->tile[y][x].substitution[s].data = dw;
                        break;

                  } // end of : switch (lay_stream[i]) {
               } // end of : if ((x < max_x) && (y < max_y))
            } // end of : if (remain_bytes >= 4) {
            else
            {
               // update new width & height of that map
               ds1->width  = max_x;
               ds1->height = max_y;

               // error
               glb_ds1_error = DS1_EOF;
               return ds1;
            }
         }
      }
   }

   // update new width & height of that map
   ds1->width  = max_x;
   ds1->height = max_y;

   // read objects
   new_obj_num   = 0;
   ds1->objects_num = 0;
   if (ds1->version >= 2)
   {
      // read # of objects
      if (remain_bytes >= 4)
      {
         ds1->objects_num = * lptr;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }

      // read all objects
      curr_obj      = NULL;
      max_subtile_x = ds1->width  * 5;
      max_subtile_y = ds1->height * 5;
      for (i=0; i < ds1->objects_num; i++)
      {
         // init this object
         memset( & obj, 0, sizeof(OBJECT_S));

         // read type
         if (remain_bytes >= 4)
         {
            obj.type = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read id
         if (remain_bytes >= 4)
         {
            obj.id = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read x
         if (remain_bytes >= 4)
         {
            obj.x = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read y
         if (remain_bytes >= 4)
         {
            obj.y = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read flags
         if (ds1->version > 5)
         {
            if (remain_bytes >= 4)
            {
               obj.flags = * lptr;
               lptr++;
               remain_bytes -= 4;
            }
            else
            {
               glb_ds1_error = DS1_EOF;
               return ds1;
            }
         }

         // is this object in the map (resizing problem) ?
         if ((obj.x < max_subtile_x) && (obj.y < max_subtile_y))
         {
            // yes, so save it

            // malloc
            save_obj = (OBJECT_S *) malloc(sizeof(OBJECT_S));
            if (save_obj == NULL)
            {
               glb_ds1_error = OBJ_MALLOC_FAILED;
               return ds1;
            }
            memset(save_obj, 0, sizeof(OBJECT_S));

            // add it to the list
            if (curr_obj == NULL)
            {
               // list is empty
               ds1->first_object = save_obj;
            }
            else
            {
               // add to the previous object
               curr_obj->next_object = save_obj;
               save_obj->prev_object = curr_obj;
            }
            curr_obj = save_obj;

            // copy its datas
            curr_obj->type  = obj.type;
            curr_obj->id    = obj.id;
            curr_obj->x     = obj.x;
            curr_obj->y     = obj.y;
            curr_obj->flags = obj.flags;

            // update object counter
            new_obj_num++;
         }
      }
   }

   // update the new number of objects in that ds1
   ds1->objects_num = new_obj_num;

   /* read substitution group datas

      warning : in fact there can be less groups than expected
      like in data\global\tiles\act1\outdoors\trees.ds1, where the file
      stop right after the last x0 data, leaving the other
      datas unknown (y0, width, height), and npc paths unknown */

   if ( (ds1->version >= 12)  &&
        ((ds1->substitution_type == 1) || (ds1->substitution_type == 2))
      )
   {
      // skip 1 dword ?
      if (ds1->version >= 18)
      {
         lptr++;
         remain_bytes -= 4;
      }

      // read # of substitution groups
      if (remain_bytes >= 4)
      {
         ds1->substitution_groups_num = * lptr;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }

      // reserved memory for the substitution groups datas
      s = sizeof(SUBST_GROUP_S) * ds1->substitution_groups_num;
      ds1->group = (SUBST_GROUP_S *) malloc(s);
      if (ds1->group == NULL)
      {
         glb_ds1_error = GROUPS_MALLOC_FAILED;
         return ds1;
      }
      memset(ds1->group, 0, s);

      // read group datas
      for (i=0; i < ds1->substitution_groups_num; i++)
      {
         // read x0
         if (remain_bytes >= 4)
         {
            ds1->group[i].x0 = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read y0
         if (remain_bytes >= 4)
         {
            ds1->group[i].y0 = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read width
         if (remain_bytes >= 4)
         {
            ds1->group[i].width = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read height
         if (remain_bytes >= 4)
         {
            ds1->group[i].height = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read (optional) unknown data
         if (ds1->version >= 13)
         {
            if (remain_bytes >= 4)
            {
               ds1->group[i].unknown = * lptr;
               lptr++;
               remain_bytes -= 4;
            }
            else
            {
               glb_ds1_error = DS1_EOF;
               return ds1;
            }
         }
      }
   }

   // read paths datas
   if (ds1->version >= 14)
   {
      // read # of npc
      if (remain_bytes >= 4)
      {
         npc_num = * lptr;
         lptr++;
         remain_bytes -= 4;
      }
      else
      {
         glb_ds1_error = DS1_EOF;
         return ds1;
      }

      // for all npc
      for (i=0; i < npc_num; i++)
      {
         // read # of paths for this npc
         if (remain_bytes >= 4)
         {
            path_num = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read X coordinate for this npc
         if (remain_bytes >= 4)
         {
            lx = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // read Y coordinate for this npc
         if (remain_bytes >= 4)
         {
            ly = * lptr;
            lptr++;
            remain_bytes -= 4;
         }
         else
         {
            glb_ds1_error = DS1_EOF;
            return ds1;
         }

         // search of which object these coordinates are pointing
         curr_obj = ds1->first_object;
         found    = FALSE;
         while ((curr_obj != NULL) && (found == FALSE))
         {
            if ((curr_obj->x == lx) && (curr_obj->y == ly))
               found = TRUE;
            else
               curr_obj = curr_obj->next_object;
         }

         if ((found == TRUE) && (curr_obj->first_path == NULL))
         {
            // ok, let's read these paths datas

            // for all path of this npc
            curr_path    = NULL;
            new_path_num = 0;
            for (n=0; n < path_num; n++)
            {
               memset( & path, 0, sizeof(PATH_S));

               // read path x 
               if (remain_bytes >= 4)
               {
                  path.x = * lptr;
                  lptr++;
                  remain_bytes -= 4;
               }
               else
               {
                  glb_ds1_error = DS1_EOF;
                  return ds1;
               }

               // read path y
               if (remain_bytes >= 4)
               {
                  path.y = * lptr;
                  lptr++;
                  remain_bytes -= 4;
               }
               else
               {
                  glb_ds1_error = DS1_EOF;
                  return ds1;
               }

               // read (optional) action
               if (ds1->version >= 15)
               {
                  if (remain_bytes >= 4)
                  {
                     path.action = * lptr;
                     lptr++;
                     remain_bytes -= 4;
                  }
                  else
                  {
                     glb_ds1_error = DS1_EOF;
                     return ds1;
                  }
               }

               // reserved memory for this path
               s = sizeof(PATH_S);
               save_path = (PATH_S *) malloc(s);
               if (save_path == NULL)
               {
                  glb_ds1_error = PATH_MALLOC_FAILED;
                  return ds1;
               }
               memset(save_path, 0, s);

               // add it to the list
               if (curr_path == NULL)
               {
                  // list is empty
                  curr_obj->first_path = save_path;
               }
               else
               {
                  // add to the previous path
                  curr_path->next_path = save_path;
                  save_path->prev_path = curr_path;
               }
               curr_path = save_path;

               // copy its datas
               curr_path->x      = path.x;
               curr_path->y      = path.y;
               curr_path->action = path.action;

               // update paths counter
               new_path_num++;
            }

            // update the new number of paths in that object
            curr_obj->path_num = new_path_num;
         }
         else
         {
            // no object at these coordinates, or object already has paths

            // skip paths infos
            if (ds1->version >= 15)
            {
               // skip X, Y and Action
               s = path_num * 3;
               lptr += s;
               remain_bytes -= (s * 4);
            }
            else
            {
               // skip X and Y only, no Action
               s = path_num * 2;
               lptr += s;
               remain_bytes -= (s * 4);
            }
         }
      }
   }

   // end
   return ds1;

}


// ====================================================================================
// save a ds1 on disk
// ====================================================================================
int ds1_save(char * name, DS1_S * ds1)
{
   OBJECT_S * curr_obj;
   PATH_S   * curr_path;
   FILE     * out;
   int      i, x, y, p;
   long     dw, npc;


   out = fopen(name, "wb");
   if (out == NULL)
      return -1;

   dw = 18;
   fwrite( & dw, 4, 1, out);

   dw = ds1->width - 1;
   fwrite( & dw, 4, 1, out);

   dw = ds1->height - 1;
   fwrite( & dw, 4, 1, out);

   dw = ds1->act - 1;
   fwrite( & dw, 4, 1, out);

   dw = ds1->substitution_type;
   fwrite( & dw, 4, 1, out);

   dw = ds1->file_num;
   fwrite( & dw, 4, 1, out);

   for (i=0; i < ds1->file_num; i++)
      fwrite(ds1->file_ptr[i], strlen(ds1->file_ptr[i]) + 1, 1, out);

   dw = ds1->wall_num;
   fwrite( & dw, 4, 1, out);

   dw = ds1->floor_num;
   fwrite( & dw, 4, 1, out);

   // walls
   for (i=0; i < ds1->wall_num; i++)
   {
      // props
      for (y=0; y < ds1->height; y++)
      {
         for (x=0; x < ds1->width; x++)
         {
            dw  = ds1->tile[y][x].wall[i].prop1;
            dw |= (ds1->tile[y][x].wall[i].sub_index  << 8);
            dw |= (ds1->tile[y][x].wall[i].unk1       << 14);
            dw |= (ds1->tile[y][x].wall[i].main_index << 20);
            dw |= (ds1->tile[y][x].wall[i].unk2       << 26);
            dw |= (ds1->tile[y][x].wall[i].hidden     << 31);
            fwrite( & dw, 4, 1, out);
         }
      }

      // orientation
      for (y=0; y < ds1->height; y++)
      {
         for (x=0; x < ds1->width; x++)
         {
            dw  = ds1->tile[y][x].wall[i].orientation;
            dw |= ds1->tile[y][x].wall[i].zero;
            fwrite( & dw, 4, 1, out);
         }
      }
   }
   
   // floors
   for (i=0; i < ds1->floor_num; i++)
   {
      // props
      for (y=0; y < ds1->height; y++)
      {
         for (x=0; x < ds1->width; x++)
         {
            dw  = ds1->tile[y][x].floor[i].prop1;
            dw |= (ds1->tile[y][x].floor[i].sub_index  << 8);
            dw |= (ds1->tile[y][x].floor[i].unk1       << 14);
            dw |= (ds1->tile[y][x].floor[i].main_index << 20);
            dw |= (ds1->tile[y][x].floor[i].unk2       << 26);
            dw |= (ds1->tile[y][x].floor[i].hidden     << 31);
            fwrite( & dw, 4, 1, out);
         }
      }
   }

   // shadows
   for (i=0; i < ds1->shadow_num; i++)
   {
      // props
      for (y=0; y < ds1->height; y++)
      {
         for (x=0; x < ds1->width; x++)
         {
            dw  = ds1->tile[y][x].shadow[i].prop1;
            dw |= (ds1->tile[y][x].shadow[i].sub_index  << 8);
            dw |= (ds1->tile[y][x].shadow[i].unk1       << 14);
            dw |= (ds1->tile[y][x].shadow[i].main_index << 20);
            dw |= (ds1->tile[y][x].shadow[i].unk2       << 26);
            dw |= (ds1->tile[y][x].shadow[i].hidden     << 31);
            fwrite( & dw, 4, 1, out);
         }
      }
   }

   // (optional) substitution tiles
   if (ds1->substitution_type)
   {
      for (i=0; i < ds1->substitution_num; i++)
      {
         // props
         for (y=0; y < ds1->height; y++)
         {
            for (x=0; x < ds1->width; x++)
            {
               dw = ds1->tile[y][x].substitution[i].data;
               fwrite( & dw, 4, 1, out);
            }
         }
      }
   }

   // objects
   fwrite( & ds1->objects_num, 4, 1, out);
   curr_obj = ds1->first_object;
   npc = 0;
   for (i=0; i < ds1->objects_num; i++)
   {
      fwrite( & curr_obj->type,  4, 1, out);
      fwrite( & curr_obj->id,    4, 1, out);
      fwrite( & curr_obj->x,     4, 1, out);
      fwrite( & curr_obj->y,     4, 1, out);
      fwrite( & curr_obj->flags, 4, 1, out);
      if (curr_obj->first_path != NULL)
         npc++;
      curr_obj = curr_obj->next_object;
   }

   // optional groups
   if (ds1->substitution_type )
   {
      // put a 0 dword
      dw = 0;
      fwrite( & dw, 4, 1, out);

      // # of groups
      fwrite( & ds1->substitution_groups_num, 4, 1, out);

      // groups
      for (i=0; i < ds1->substitution_groups_num; i++)
      {
         fwrite( & ds1->group[i].x0,      4, 1, out);
         fwrite( & ds1->group[i].y0,      4, 1, out);
         fwrite( & ds1->group[i].width,   4, 1, out);
         fwrite( & ds1->group[i].height,  4, 1, out);
         fwrite( & ds1->group[i].unknown, 4, 1, out);
      }
   }

   // npc paths
   fwrite( & npc, 4, 1, out);
   curr_obj = ds1->first_object;
   for (i=0; i < npc; i++)
   {
      while (curr_obj->first_path == NULL)
         curr_obj = curr_obj->next_object;
      fwrite( & curr_obj->path_num, 4, 1, out);
      fwrite( & curr_obj->x,        4, 1, out);
      fwrite( & curr_obj->y,        4, 1, out);
      curr_path = curr_obj->first_path;
      for (p=0; p < curr_obj->path_num; p++)
      {
         fwrite( & curr_path->x,      4, 1, out);
         fwrite( & curr_path->y,      4, 1, out);
         fwrite( & curr_path->action, 4, 1, out);
         curr_path = curr_path->next_path;
      }
      curr_obj = curr_obj->next_object;
   }

   // end
   fclose(out);
   return 0;
}


// ====================================================================================
int main(int argc, char ** argv)
{
   char  * buffer;
   long  buffer_length;
   DS1_S * ds1;
   int   ret, cell_x, cell_y;


   if (argc != 4)
   {
      printf("syntaxe : ds1.exe <file.ds1> <cell X> <cell Y>\n");
      return -1;
   }

   printf("sizeof(TILE_S) = %i\n", sizeof (TILE_S));
   printf("a map of 250*250 tiles would be %i bytes (%i kB, or %i MB)\n",
      sizeof (TILE_S) * 250*250,
      sizeof (TILE_S) * 250 * 250 / 1024,
      sizeof (TILE_S) * 250 * 250 / 1024 / 1024
   );

   buffer = ds1_disk_to_mem(argv[1], & buffer_length);
   if (buffer != NULL)
   {
      ds1 = ds1_load(buffer, buffer_length, -1, -1);
      printf("glb_ds1_error = %i\n\n", glb_ds1_error);

      printf("version                 = %li\n", ds1->version);
      printf("act                     = %li\n", ds1->act);
      printf("file_num                = %li\n", ds1->file_num);
      printf("floor_num               = %i\n",  ds1->floor_num);
      printf("wall_num                = %i\n",  ds1->wall_num);
      printf("shadow_num              = %i\n",  ds1->shadow_num);
      printf("substitution_num        = %i\n",  ds1->substitution_num);
      printf("substitution_groups_num = %i\n",  ds1->substitution_groups_num);
      printf("substitution_type       = %li\n", ds1->substitution_type);
      printf("objects_num             = %i\n",  ds1->objects_num);
      printf("width                   = %li\n", ds1->width);
      printf("height                  = %li\n", ds1->height);
      printf("original_width          = %li\n", ds1->original_width);
      printf("original_height         = %li\n", ds1->original_height);

      cell_x = atoi(argv[2]);
      cell_y = atoi(argv[3]);
      printf("\ncell (%i, %i) :\n===============\n", cell_x, cell_y);
      printf("floor[0].main_index     = %i\n",   ds1->tile[cell_y][cell_x].floor[0].main_index);
      printf("floor[0].sub_index      = %i\n",   ds1->tile[cell_y][cell_x].floor[0].sub_index);
      printf("floor[0].prop1          = %i\n\n", ds1->tile[cell_y][cell_x].floor[0].prop1);

      printf("floor[1].main_index     = %i\n",   ds1->tile[cell_y][cell_x].floor[1].main_index);
      printf("floor[1].sub_index      = %i\n",   ds1->tile[cell_y][cell_x].floor[1].sub_index);
      printf("floor[1].prop1          = %i\n\n", ds1->tile[cell_y][cell_x].floor[1].prop1);

      printf("wall[0].orientation    = %i\n",   ds1->tile[cell_y][cell_x].wall[0].orientation);
      printf("wall[0].main_index     = %i\n",   ds1->tile[cell_y][cell_x].wall[0].main_index);
      printf("wall[0].sub_index      = %i\n",   ds1->tile[cell_y][cell_x].wall[0].sub_index);
      printf("wall[0].prop1          = %i\n\n", ds1->tile[cell_y][cell_x].wall[0].prop1);

      printf("wall[1].orientation    = %i\n",   ds1->tile[cell_y][cell_x].wall[1].orientation);
      printf("wall[1].main_index     = %i\n",   ds1->tile[cell_y][cell_x].wall[1].main_index);
      printf("wall[1].sub_index      = %i\n",   ds1->tile[cell_y][cell_x].wall[1].sub_index);
      printf("wall[1].prop1          = %i\n\n", ds1->tile[cell_y][cell_x].wall[1].prop1);


      printf("wall[2].orientation    = %i\n",   ds1->tile[cell_y][cell_x].wall[2].orientation);
      printf("wall[2].main_index     = %i\n",   ds1->tile[cell_y][cell_x].wall[2].main_index);
      printf("wall[2].sub_index      = %i\n",   ds1->tile[cell_y][cell_x].wall[2].sub_index);
      printf("wall[2].prop1          = %i\n\n", ds1->tile[cell_y][cell_x].wall[2].prop1);


      printf("wall[3].orientation    = %i\n",   ds1->tile[cell_y][cell_x].wall[3].orientation);
      printf("wall[3].main_index     = %i\n",   ds1->tile[cell_y][cell_x].wall[3].main_index);
      printf("wall[3].sub_index      = %i\n",   ds1->tile[cell_y][cell_x].wall[3].sub_index);
      printf("wall[3].prop1          = %i\n\n", ds1->tile[cell_y][cell_x].wall[3].prop1);
      fflush(stdout);

      ret = ds1_save("new_map.ds1", ds1);
      printf("\nds1_save return = %i\n", ret);
      fflush(stdout);

      ds1_destroy(ds1);
      free(buffer);
   }
   else
   {
      printf("can't read <%s>\n", argv[1]);
      return -1;
   }

   // end
   return 0;
}
